/* das08jr
 * =======
 *
 * Linux device driver for Computer Boards CIO-DAS08/Jr-AO board.
 *
 * This board has 8 analog input lines and 2 analog output lines. Additionally
 * there are 8 digital outputs.
 * These lines are mapped onto minor 0 (digital output), 1 - 8 for analog
 * input, and 1 + 2 for analog output by this driver.
 *
 * Copyright (C) 1998,1999 Jochen Kpper
 * Copyright (C) 1999 John L. Bergmans
 */


#define __KERNEL__
#define MODULE


#include <linux/config.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/version.h>

#include <asm/io.h>
#include <asm/system.h>
#include <asm/uaccess.h>

#include "das08jr.h"


/* port addresses to autoprobe
 * If the parameter io=0xNNN is not given at load time, these addresses are
 * probed in their respective order until one device is found or all specified
 * addresses are probed */
static unsigned long autoprobe_port[] =
{
	0x378, /* this is where we are, remove for distribution */
	0x200,
	0x270,
	0x2e0,
	0x2e8,
	0x2f0,
	0x2f8,
	0x300,
	0x310,
	0x320,
	0x378,
	0
};



/* Module Name */
#define DAS08JR_NAME "das08jr"


/* driver version */
static const char das08jr_version[] = "$Date: 1999/10/25 09:21:41 $";


/* user configurable port base
 * If you give a parameter io=0xNNN, this address will be used,
 * autoprobing will be done otherwise. */
unsigned long io = 0;
MODULE_PARM( io, "l" );


/* number of available lines */
#define DAS08JR_ANALOG_INPUTS    ( 8 )
#define DAS08JR_ANALOG_OUTPUTS   ( 2 )
#define DAS08JR_DIGITAL_LINES    ( 1 )
#define DAS08JR_MAX_LINE         ( 8 ) /* we have nine lines starting at zero */

/* number of ports */
#define DAS08JR_SIZE             ( 8 )

/* Port Offsets */
#define DAS08JR_ANALOG_IN_LSB    ( 0 ) /* read */
#define DAS08JR_ANALOG_IN_MSB    ( 1 ) /* read */
#define DAS08JR_START            ( 1 )
#define DAS08JR_STATUS           ( 2 ) /* read */
#define DAS08JR_MUX              ( 2 ) /* read */
#define DAS08JR_CONTROL          ( 2 )
#define DAS08JR_DIGITAL_IN       ( 3 ) /* read */
#define DAS08JR_DIGITAL_OUT      ( 3 )
#define DAS08JR_ANALOG_OUT_LSB0  ( 4 )
#define DAS08JR_ANALOG_OUT_MSB0  ( 5 )
#define DAS08JR_ANALOG_OUT_LSB1  ( 6 )
#define DAS08JR_ANALOG_OUT_MSB1  ( 7 )



/* local functions */
static unsigned long    das08jr_check_port( unsigned long );
static int              das08jr_open( struct inode *, struct file * );
static int              das08jr_release( struct inode *, struct file * );
static ssize_t          das08jr_read( struct file *, char *, size_t, loff_t * );
static ssize_t          das08jr_read_digital( struct file *, char *, size_t, loff_t * );
static ssize_t          das08jr_write_digital( struct file *, const char *, size_t, loff_t * );
static ssize_t          das08jr_write_ro( struct file *, const char *, size_t, loff_t * );
static ssize_t          das08jr_write_rw( struct file *, const char *, size_t, loff_t * );




/* implemented file operations for digital lines */
static struct file_operations das08jr_fops_digital =
{
    read:    das08jr_read_digital,
    write:   das08jr_write_digital,
    open:    das08jr_open,
    release: das08jr_release
};


/* implemented file operations for readonly analog lines ( 3 - 8 ) */
static struct file_operations das08jr_fops_ro =
{
    read:    das08jr_read,
    write:   das08jr_write_ro,
    open:    das08jr_open,
    release: das08jr_release
};


/* implemented file operations for writable analog lines( 1 + 2 ) */
static struct file_operations das08jr_fops_rw =
{
    read:    das08jr_read,
    write:   das08jr_write_rw,
    open:    das08jr_open,
    release: das08jr_release
};



/**
 * Probe for device, register ports, register as character device,
 * export exactly the symbols we want to be public (none yet).
 *
 * @return Success status, zero on success.
 */
int init_module( void )
{
	/* hide all symbols */
	EXPORT_NO_SYMBOLS;

	/* check device port base and request the port region */
    if( 0 == io ) {
		int i = 0;
		while ( 0 != ( io = autoprobe_port[ i ] ) ) {
			if( 0 != das08jr_check_port( io ) )
				break;
			i++;
		}
		if( 0 == io ) {
			return( -ENODEV );
		}
    } else {
		if( 0 == das08jr_check_port( io ) )
			return( -ENODEV );
	}
	/* register as character device */
	if( 0 > register_chrdev( DAS08JR_MAJOR, DAS08JR_NAME, &das08jr_fops_rw ) ) {
		printk( KERN_WARNING DAS08JR_NAME ": Cannot register major number %d.\n", DAS08JR_MAJOR );
		release_region( io, DAS08JR_SIZE );
		return( -EBUSY );
	} 
	/* finish up */
	printk( KERN_INFO DAS08JR_NAME ": CIO-DAS08Jr-AO driver (%s) loaded\n"
			DAS08JR_NAME ": Copyright (C) 1998 Jochen Kuepper\n"
			DAS08JR_NAME ": found board at 0x%lx\n",
			das08jr_version, io );
    return( 0 );
}



/**
 * Cleanup when we are unloaded.
 */ 
void cleanup_module( void )
{
    unregister_chrdev( DAS08JR_MAJOR, DAS08JR_NAME );
    release_region( io, DAS08JR_SIZE );
    printk( KERN_INFO DAS08JR_NAME " unloaded\n" );
    return;
}



/**
 * Check the given port base address.
 *
 * @param base Port address to check for a device at.
 * @return     The port base if we found a conforming device, 0 otherwise.
 */ 
static unsigned long das08jr_check_port( unsigned long base )
{
	/* check and register given port region */
	if( -EBUSY == check_region( base, DAS08JR_SIZE ) )
		return 0;
	request_region( base, DAS08JR_SIZE, DAS08JR_NAME );
	/* Try out the port */
	outb_p( 0x03, base + DAS08JR_CONTROL );
	if( 0x03 != ( 0x07 & inb_p( base + DAS08JR_MUX ) ) ) {
		printk( KERN_DEBUG DAS08JR_NAME ": no device at %lx\n", base );
		release_region( base, DAS08JR_SIZE );
		return 0;
	}
	outb_p( 0x06, base + DAS08JR_CONTROL );
	if( 0x06 != ( 0x07 & inb_p( base + DAS08JR_MUX ) ) ) {
		printk( KERN_DEBUG DAS08JR_NAME ": no device at %lx\n", base );
		release_region( base, DAS08JR_SIZE );
		return 0;
	}
	return( base );
}



/**
 * Open device.
 */ 
static int das08jr_open( struct inode *inode, struct file *fp )
{
    int minor = MINOR( inode->i_rdev );

	if( DAS08JR_MAX_LINE < minor )
		return( -ENODEV );

    if( DAS08JR_DIGITAL_LINES > minor ) {
		/* digital lines */
		fp->f_op = &das08jr_fops_digital;
	} else if( DAS08JR_ANALOG_OUTPUTS + DAS08JR_DIGITAL_LINES > minor ) {
		/* read/writeable analog lines */
		fp->f_op = &das08jr_fops_rw;
	} else {
		/* readonly analog lines */
		if( FMODE_WRITE & fp->f_mode ) {
			printk( KERN_WARNING DAS08JR_NAME "Access denied\n" );
			return( -EACCES );
		}
		fp->f_op = &das08jr_fops_ro;
	}
	
    MOD_INC_USE_COUNT;
    return( 0 );
}



/**
 * Close device
 */ 
static int das08jr_release( struct inode *inode, struct file *fp )
{
    MOD_DEC_USE_COUNT;
    return( 0 );
}




/**
 * Read digital lines
 */
static int das08jr_read_digital( struct file *fp, char *buf, size_t count, loff_t *off )
{
	if( count > 0 ) {
		unsigned char in = inb( io + DAS08JR_DIGITAL_IN );
		put_user( in, buf );
		return( 1 );		
	}
	return( 0 );
}




/**
 * write digital lines
 */
static int das08jr_write_digital( struct file *fp, const char *buf, size_t count, loff_t *off )
{
	if( count > 0 ) {
		unsigned char out;
		get_user( out, buf );
		outb( out, io + DAS08JR_DIGITAL_OUT );
		return( 1 );
	}
	return( 0 );
}



/**
 * Read from analog line
 * we always read one short value !
 */
static int das08jr_read( struct file *file, char *buf, size_t count, loff_t *off )
{  
	unsigned int channel = MINOR( file->f_dentry->d_inode->i_rdev ) - 1;
	/* Mapping of channel number to minor number is as follows:
	   Channel:  Minor:
	   0        1
	   1        2
	   ...      ...
	   7        8
	*/
	outb_p( channel, io + DAS08JR_CONTROL );      // Select channel
	outb_p( 0xff, io + DAS08JR_START );           // Trigger conversion
	while( 0x80 & inb_p( io + DAS08JR_STATUS ) )  // Loop until EOC clear
		;
	if( count >= 2 ) { 
		/* read data and put it into user space */
		unsigned char lsb;
		unsigned char msb;
		lsb = inb_p( io + DAS08JR_ANALOG_IN_LSB );
		msb = inb_p( io + DAS08JR_ANALOG_IN_MSB );
		put_user( ( ((int)msb) << 4 ) + ( lsb >> 4 ), (unsigned short *)buf );
		return( 2 );
	}
    return( 0 );
}



/**
 * Write to writeable analog line.
 *
 * We do not need to check wether this really is an writeable line, since open
 * cares about that an we only get called on lines 1 or 2.
 *
 * We simply put one value to the output ports, start D/A conversion, and return.
 *
 * @return On success the number of written bytes ( always 2 ),
 *         0 if we haven't gotten enough data,
 *         a negative error code otherwise.
 */
static int das08jr_write_rw( struct file *file, const char *buf, size_t count, loff_t *off )
{
    /* we need a full short value */
    if( count >= 2 ) {
		unsigned short val;
		get_user( val, (unsigned short *)buf );
		if( DAS08JR_ANALOG_OUTPUTS == MINOR( file->f_dentry->d_inode->i_rdev ) ) {
			outb( (unsigned char)( val >> 8 ), io + DAS08JR_ANALOG_OUT_MSB1 );
			outb( (unsigned char)( val & 0xff ), io + DAS08JR_ANALOG_OUT_LSB1 );
		} else {
			outb( (unsigned char)( val >> 8 ), io + DAS08JR_ANALOG_OUT_MSB0 );
			outb( (unsigned char)( val & 0xff ), io + DAS08JR_ANALOG_OUT_LSB0 );
		}
		/* Read DIO control register to initiate DAC */
		inb_p( io + DAS08JR_DIGITAL_IN);
		return( 2 );
    }
    return( 0 );
}



/**
 * Write to readonly analog line.
 *
 * @return Since we are always called on readonly lines, simply return -EPERM.
 */
static int das08jr_write_ro( struct file *fp, const char *buf, size_t count, loff_t *off )
{
    return( -EPERM );
}





//* Local Variables:
//* mode: C
//* c-file-style: "Stroustrup"
//* End:
